from collections import namedtuple

import numpy
import pandas
import pytest
from fbprophet import Prophet
from pmdarima import AutoARIMA

from hts.model import AutoArimaModel, FBProphetModel, HoltWintersModel, SarimaxModel
from hts.model.base import TimeSeriesModel


def test_instantiate_fb_model_uv(uv_tree):
    fb = FBProphetModel(node=uv_tree)
    assert isinstance(fb, TimeSeriesModel)
    fb = FBProphetModel(node=uv_tree, capacity_max=1)
    assert isinstance(fb, TimeSeriesModel)
    fb = FBProphetModel(node=uv_tree, capacity_min=1)
    assert isinstance(fb, TimeSeriesModel)


def test_fit_predict_fb_model_mv(mv_tree):
    exog = pandas.DataFrame({"precipitation": [1], "temp": [20]})
    fb = FBProphetModel(node=mv_tree)
    assert isinstance(fb, TimeSeriesModel)
    fb.fit()
    fb.predict(mv_tree, exogenous_df=exog)
    assert isinstance(fb.forecast, pandas.DataFrame)
    assert isinstance(fb.residual, numpy.ndarray)
    assert isinstance(fb.mse, float)


def test_fit_predict_fb_model_uv(uv_tree):
    fb = FBProphetModel(node=uv_tree)
    fb.fit()
    assert isinstance(fb.model, Prophet)
    fb.predict(uv_tree)
    assert isinstance(fb.forecast, pandas.DataFrame)
    assert isinstance(fb.residual, numpy.ndarray)
    assert isinstance(fb.mse, float)


def test_fit_predict_ar_model_mv(mv_tree):
    ar = AutoArimaModel(node=mv_tree)
    ar.fit(max_iter=1)
    assert isinstance(ar.model, AutoARIMA)
    exog = pandas.DataFrame({"precipitation": [1], "temp": [20]})
    ar.predict(mv_tree, steps_ahead=1, exogenous_df=exog)
    assert isinstance(ar.forecast, pandas.DataFrame)
    assert isinstance(ar.residual, numpy.ndarray)
    assert isinstance(ar.mse, float)


def test_fit_predict_ar_model_uv(uv_tree):
    ar = AutoArimaModel(node=uv_tree,)
    ar.fit(max_iter=1)
    assert isinstance(ar.model, AutoARIMA)
    ar.predict(uv_tree)
    assert isinstance(ar.forecast, pandas.DataFrame)
    assert isinstance(ar.residual, numpy.ndarray)
    assert isinstance(ar.mse, float)


def test_fit_predict_sarimax_model_uv(uv_tree):
    sar = SarimaxModel(node=uv_tree, max_iter=1,)
    fitted_sar = sar.fit()
    assert isinstance(fitted_sar, SarimaxModel)
    sar.predict(uv_tree)
    assert isinstance(sar.forecast, pandas.DataFrame)
    assert isinstance(sar.residual, numpy.ndarray)
    assert isinstance(sar.mse, float)


def test_fit_predict_hw_model_uv(uv_tree):
    hw = HoltWintersModel(node=uv_tree,)
    fitted_hw = hw.fit()
    assert isinstance(fitted_hw, HoltWintersModel)
    hw.predict(uv_tree)
    assert isinstance(hw.forecast, pandas.DataFrame)
    assert isinstance(hw.residual, numpy.ndarray)
    assert isinstance(hw.mse, float)


def test_fit_predict_hw_model_uv_with_transform(uv_tree):
    Transform = namedtuple("Transform", ["func", "inv_func"])
    transform_pos_neg = Transform(func=numpy.exp, inv_func=lambda x: -x)

    hw = HoltWintersModel(node=uv_tree, transform=transform_pos_neg)
    fitted_hw = hw.fit()
    assert isinstance(fitted_hw, HoltWintersModel)
    preds = hw.predict(uv_tree)
    assert not (preds.forecast.values > 0).any()

    assert isinstance(hw.forecast, pandas.DataFrame)
    assert isinstance(hw.residual, numpy.ndarray)
    assert isinstance(hw.mse, float)


def test_fit_predict_model_invalid_transform(uv_tree):
    Transform = namedtuple("Transform", ["func_invalid_arg", "inv_func"])
    transform_pos_neg = Transform(func_invalid_arg=numpy.exp, inv_func=lambda x: -x)
    with pytest.raises(ValueError):
        HoltWintersModel(node=uv_tree, transform=transform_pos_neg)



from typing import Dict, Optional, Union

import numpy
import pandas

from hts._t import ArrayLike, MethodT, NAryTreeT, TransformT
from hts.functions import to_sum_mat
from hts.revision import RevisionMethod


def _to_numpy(v: ArrayLike, kind: str = "forecasts") -> numpy.ndarray:
    if isinstance(v, numpy.ndarray):
        if v.ndim > 1:
            raise ValueError(f"`{kind}` values must be of a dimension 1")
        return v
    elif isinstance(v, pandas.DataFrame):
        if len(v.columns) > 1:
            raise ValueError(
                f"If providing `{kind}` as a DataFrame, it must have one column only"
            )
        col_name = v.columns[0]
        return v[col_name].values
    elif isinstance(v, pandas.Series):
        return v.values
    else:
        raise ValueError(
            f"`{kind}` must be a dict mapping string to array, series or DataFrame"
        )


def _sanitize_errors_dict(errors: Dict[str, float]) -> Dict[str, float]:
    for k, v in errors.items():
        if not isinstance(v, float):
            raise ValueError("`errors` dict must be a mapping from string to float")
    return errors


def _sanitize_residuals_dict(
    residuals: Dict[str, ArrayLike]
) -> Dict[str, numpy.ndarray]:
    for k, v in residuals.items():
        residuals[k] = _to_numpy(v, kind="residuals")
    return residuals


def _sanitize_forecasts_dict(
    forecasts: Dict[str, ArrayLike]
) -> Dict[str, pandas.DataFrame]:

    for k, v in forecasts.items():
        as_array = _to_numpy(v, kind="forecasts")
        forecasts[k] = pandas.DataFrame({"yhat": as_array})
    return forecasts


def _calculate_errors(
    method: str,
    errors: Optional[Dict[str, float]] = None,
    residuals: Optional[Dict[str, numpy.ndarray]] = None,
):
    errors_or_residuals = (
        True if (errors is not None or residuals is not None) else False
    )
    if not errors_or_residuals:
        raise ValueError(
            f"Method {method} requires either errors or residuals to be provided"
        )
    if residuals is not None:
        residuals = _sanitize_residuals_dict(residuals)
        if errors is None:
            errors = {}
        for k, v in residuals.items():
            errors[k] = numpy.mean(numpy.array(v) ** 2)
    return _sanitize_errors_dict(errors)


def revise_forecasts(
    method: str,
    forecasts: Dict[str, ArrayLike],
    errors: Optional[Dict[str, float]] = None,
    residuals: Optional[Dict[str, ArrayLike]] = None,
    summing_matrix: numpy.ndarray = None,
    nodes: NAryTreeT = None,
    transformer: TransformT = None,
):
    """
    Convenience function to get revised forecast for pre-computed base forecasts

    Parameters
    ----------
    method : str
        The reconciliation method to use
    forecasts : Dict[str, ArrayLike]
        A dict mapping key name to its forecasts (including in-sample forecasts). Required, can be
        of type ``numpy.ndarray`` of ``ndim == 1``, ``pandas.Series``, or single columned ``pandas.DataFrame``
    errors : Dict[str, float]
        A dict mapping key name to the in-sample errors. Required for methods: ``OLS``, ``WLSS``, ``WLSV`` if
        ``residuals`` is not passed
    residuals : Dict[str, ArrayLike]
        A dict mapping key name to the residuals of in-sample forecasts. Required for methods: ``OLS``, ``WLSS``,
        ``WLSV``, can be of type ``numpy.ndarray`` of ndim == 1, ``pandas.Series``, or single columned
        ``pandas.DataFrame``. If passing residuals, ``errors`` dict is not required and will instead be calculated
        using MSE metric: ``numpy.mean(numpy.array(residual) ** 2)``
    summing_matrix : numpy.ndarray
        Not required if ``nodes`` argument is passed, or if using ``BU`` approach
    nodes : NAryTreeT
        The tree of nodes as specified in :py:class:`HierarchyTree <hts.hierarchy.HierarchyTree>`. Required if not
        if using ``AHP``, ``PHA``, ``FP`` methods, or if using  passing the ``OLS``, ``WLSS``, ``WLSV`` methods
        and not passing the ``summing_matrix`` parameter
    transformer : TransformT
        A transform with the method: ``inv_func`` that will be applied to the forecasts

    Returns
    -------
    revised forecasts : ``pandas.DataFrame``
        The revised forecasts
    """

    if nodes:
        summing_matrix, sum_mat_labels = to_sum_mat(nodes)

    if method in [MethodT.AHP.name, MethodT.PHA.name, MethodT.FP.name] and not nodes:
        raise ValueError(f"Method {method} requires an NAryTree to be passed")

    if method in [MethodT.OLS.name, MethodT.WLSS.name, MethodT.WLSV.name]:
        errors = _calculate_errors(method=method, errors=errors, residuals=residuals)
        if not (all([forecasts, errors]) or (not summing_matrix)):
            raise ValueError(
                f"Method {method} requires forecasts, errors, and residuals to be passed, as "
                f"well as an NAryTree or a summing matrix"
            )

    revision = RevisionMethod(
        name=method, sum_mat=summing_matrix, transformer=transformer
    )
    sanitized_forecasts = _sanitize_forecasts_dict(forecasts)
    revised = revision.revise(forecasts=sanitized_forecasts, mse=errors, nodes=nodes)

    return pandas.DataFrame(revised, columns=list(sanitized_forecasts.keys()))



import logging
import os
from io import StringIO

import numpy
import pandas

logger = logging.getLogger(__name__)

try:
    import requests
except ImportError:  # pragma: no cover
    logger.error(
        "Some loading functions might be impaired, install requests "
        "with: \npip install requests\n if you'd like to use them"
    )

MOBILITY_URL = "https://hierarchical-sample-data.s3.amazonaws.com/mobility.csv"
GEO_EVENTS_URL = "https://osf.io/v8qax/download"


def get_data_home(data_home=None):
    """
    Return the path of the scikit-hts data dir.

    This folder is used by some large dataset loaders to avoid downloading the
    data several times.

    By default the data dir is set to a folder named 'scikit_hts_data' in the
    user home folder.
    Alternatively, it can be set by the 'SCIKIT_HTS_DATA' environment
    variable or programmatically by giving an explicit folder path. The '~'
    symbol is expanded to the user home folder.
    If the folder does not already exist, it is automatically created.
    Parameters
    ----------
    data_home : str | None
        The path to scikit-hts data dir.
    """
    if data_home is None:
        data_home = os.environ.get(
            "SCIKIT_HTS_DATA", os.path.join("~", "scikit_hts_data")
        )
    data_home = os.path.expanduser(data_home)
    if not os.path.exists(data_home):
        os.makedirs(data_home)
    return data_home


def partition_column(column, n=3):
    partitioned = column.apply(
        lambda x: numpy.random.dirichlet(numpy.ones(n), size=1).ravel() * x
    ).values
    return [[i[j] for i in partitioned] for j in range(n)]


def load_hierarchical_sine_data(start, end, n=10000):
    dts = (end - start).total_seconds()
    dti = pandas.DatetimeIndex(
        [start + pandas.Timedelta(numpy.random.uniform(0, dts), "s") for _ in range(n)]
    ).sort_values()
    time = numpy.arange(0, len(dti), 0.01)
    amplitude = numpy.sin(time) * 10
    amplitude += numpy.random.normal(2 * amplitude + 2, 5)
    df = pandas.DataFrame(index=dti, data={"total": amplitude[0 : len(dti)]})
    df["a"], df["b"], df["c"] = partition_column(df.total, n=3)
    df["a_x"], df["a_y"] = partition_column(df.a, n=2)
    df["b_x"], df["b_y"] = partition_column(df.b, n=2)
    df["c_x"], df["c_y"] = partition_column(df.c, n=2)
    df["a_x_1"], df["a_x_2"] = partition_column(df.a_x, n=2)
    df["a_y_1"], df["a_y_2"] = partition_column(df.a_y, n=2)
    df["b_x_1"], df["b_x_2"] = partition_column(df.b_x, n=2)
    df["b_y_1"], df["b_y_2"] = partition_column(df.b_y, n=2)
    df["c_x_1"], df["c_x_2"] = partition_column(df.c_x, n=2)
    df["c_y_1"], df["c_y_2"] = partition_column(df.c_y, n=2)
    return df


def load_mobility_data(data_home=None):
    """
    Original dataset: https://www.kaggle.com/pronto/cycle-share-dataset
    Returns
    -------
    df : pandas.DataFrame
    """
    data_path = get_data_home(data_home)
    if "mobility.csv" not in os.listdir(data_path):
        df_string = requests.get(MOBILITY_URL).content
        df = pandas.read_csv(
            StringIO(df_string.decode("utf-8")),
            index_col="starttime",
            parse_dates=["starttime"],
        )
        df.reset_index().to_csv(os.path.join(data_path, "mobility.csv"), index=False)
        return df
    else:
        return pandas.read_csv(
            os.path.join(data_path, "mobility.csv"),
            index_col="starttime",
            parse_dates=["starttime"],
        )


def load_geo_events_data(data_home=None):
    """
    Returns
    -------
    df : pandas.DataFrame
    """
    data_path = get_data_home(data_home)
    if "power.csv" not in os.listdir(data_path):
        df_string = requests.get(GEO_EVENTS_URL).content
        df = pandas.read_csv(
            StringIO(df_string.decode("utf-8")),
            parse_dates=["event_ts"],
            index_col="event_ts",
        )
        df.reset_index().to_csv(os.path.join(data_path, "power.csv"), index=False)
        return df
    else:
        return pandas.read_csv(
            os.path.join(data_path, "power.csv"),
            parse_dates=["event_ts"],
            index_col="event_ts",
        )



import numpy
import pandas

import hts.hierarchy
from hts.functions import (
    _create_bl_str_col,
    get_agg_series,
    get_hierarchichal_df,
    to_sum_mat,
)


def test_sum_mat_uv(uv_tree):
    mat, sum_mat_labels = to_sum_mat(uv_tree)
    assert isinstance(mat, numpy.ndarray)
    shp = mat.shape
    assert shp[0] == uv_tree.num_nodes() + 1
    assert shp[1] == uv_tree.leaf_sum()


def test_sum_mat_mv(mv_tree):
    mat, sum_mat_labels = to_sum_mat(mv_tree)
    assert isinstance(mat, numpy.ndarray)
    shp = mat.shape
    assert shp[0] == mv_tree.num_nodes() + 1
    assert shp[1] == mv_tree.leaf_sum()


def test_sum_mat_hierarchical():
    hierarchy = {"total": ["A", "B"], "A": ["A_X", "A_Y", "A_Z"], "B": ["B_X", "B_Y"]}
    hier_df = pandas.DataFrame(
        data={
            "total": [],
            "A": [],
            "B": [],
            "A_X": [],
            "A_Y": [],
            "A_Z": [],
            "B_X": [],
            "B_Y": [],
        }
    )

    tree = hts.hierarchy.HierarchyTree.from_nodes(hierarchy, hier_df)
    sum_mat, sum_mat_labels = to_sum_mat(tree)

    expected_sum_mat = numpy.array(
        [
            [1, 1, 1, 1, 1],  # total
            [0, 0, 0, 1, 1],  # B
            [1, 1, 1, 0, 0],  # A
            [1, 0, 0, 0, 0],  # A_X
            [0, 1, 0, 0, 0],  # A_Y
            [0, 0, 1, 0, 0],  # A_Z
            [0, 0, 0, 1, 0],  # B_X
            [0, 0, 0, 0, 1],
        ]
    )  # B_Y

    numpy.testing.assert_array_equal(sum_mat, expected_sum_mat)
    assert sum_mat_labels == ["total", "B", "A", "A_X", "A_Y", "A_Z", "B_X", "B_Y"]


def test_sum_mat_grouped():
    hierarchy = {
        "total": ["A", "B", "X", "Y"],
        "A": ["A_X", "A_Y"],
        "B": ["B_X", "B_Y"],
    }
    grouped_df = pandas.DataFrame(
        data={
            "total": [],
            "A": [],
            "B": [],
            "X": [],
            "Y": [],
            "A_X": [],
            "A_Y": [],
            "B_X": [],
            "B_Y": [],
        }
    )

    tree = hts.hierarchy.HierarchyTree.from_nodes(hierarchy, grouped_df)
    sum_mat, sum_mat_labels = to_sum_mat(tree)

    expected_sum_mat = numpy.array(
        [
            [1, 1, 1, 1],  # total
            [0, 1, 0, 1],  # Y
            [1, 0, 1, 0],  # X
            [0, 0, 1, 1],  # B
            [1, 1, 0, 0],  # A
            [1, 0, 0, 0],  # A_X
            [0, 1, 0, 0],  # A_Y
            [0, 0, 1, 0],  # B_X
            [0, 0, 0, 1],  # B_Y
        ]
    )

    numpy.testing.assert_array_equal(sum_mat, expected_sum_mat)
    assert sum_mat_labels == ["total", "Y", "X", "B", "A", "A_X", "A_Y", "B_X", "B_Y"]


def test_sum_mat_visnights_hier(visnights_hier):
    hier_df = pandas.DataFrame(
        data={
            "total": [],
            "VIC": [],
            "QLD": [],
            "SAU": [],
            "WAU": [],
            "OTH": [],
            "NSW": [],
            "NSW_Metro": [],
            "NSW_NthCo": [],
            "NSW_NthIn": [],
            "NSW_SthCo": [],
            "NSW_SthIn": [],
            "OTH_Metro": [],
            "OTH_NoMet": [],
            "QLD_Cntrl": [],
            "QLD_Metro": [],
            "QLD_NthCo": [],
            "SAU_Coast": [],
            "SAU_Inner": [],
            "SAU_Metro": [],
            "VIC_EstCo": [],
            "VIC_Inner": [],
            "VIC_Metro": [],
            "VIC_WstCo": [],
            "WAU_Coast": [],
            "WAU_Inner": [],
            "WAU_Metro": [],
        }
    )

    tree = hts.hierarchy.HierarchyTree.from_nodes(visnights_hier, hier_df)
    sum_mat, sum_mat_labels = to_sum_mat(tree)

    expected_sum_mat = numpy.array(
        [
            [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],  # total
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1],  # VIC
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0],  # QLD
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0],  # SAU
            [0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # WAU
            [0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # OTH
            [1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # NSW
            [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # NSW_Metro
            [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # NSW_NthCo
            [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # NSW_NthIn
            [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # NSW_SthCo
            [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # NSW_SthIn
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # OTH_Metro
            [0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # OTH_NoMet
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # WAU_Coast
            [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # WAU_Inner
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # WAU_Metro
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],  # SAU_Coast
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],  # SAU_Inner
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],  # SAU_Metro
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],  # QLD_Cntrl
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],  # QLD_Metro
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],  # QLD_NthCo
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],  # VIC_EstCo
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],  # VIC_Inner
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],  # VIC_Metro
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],  # VIC_WstCo
        ]
    )

    numpy.testing.assert_array_equal(sum_mat, expected_sum_mat)


def test_demo_unique_constraint():
    # Example https://otexts.com/fpp2/hts.html
    # Does not work when you have elements that are named the same, but represent
    # different levels in the hierarchy. See expected_sum_mat below for example.
    hierarchy = {"total": ["A", "B"], "A": ["AA", "AB", "AC"], "B": ["BA", "BB"]}
    hier_df = pandas.DataFrame(
        data={
            "total": [],
            "A": [],
            "B": [],
            "AA": [],
            "AB": [],
            "AC": [],
            "BA": [],
            "BB": [],
        }
    )

    tree = hts.hierarchy.HierarchyTree.from_nodes(hierarchy, hier_df)
    sum_mat, sum_mat_labels = to_sum_mat(tree)

    expected_sum_mat = numpy.array(
        [
            [1, 1, 1, 1, 1],  # total
            [0, 1, 0, 1, 1],  # B, Incorrectly finds B in AB
            [1, 1, 1, 1, 0],  # A, Incorrectly finds A in BA
            [1, 0, 0, 0, 0],  # AA
            [0, 1, 0, 0, 0],  # AB
            [0, 0, 1, 0, 0],  # AC
            [0, 0, 0, 1, 0],  # BA
            [0, 0, 0, 0, 1],  # BB
        ]
    )

    numpy.testing.assert_array_equal(sum_mat, expected_sum_mat)


def test_1lev():
    grouped_df = pandas.DataFrame(
        data={"lev1": ["A", "A", "B", "B"], "lev2": ["X", "Y", "X", "Y"],}
    )

    levels = get_agg_series(grouped_df, [["lev1"]])
    expected_levels = ["A", "B"]
    assert sorted(levels) == sorted(expected_levels)

    levels = get_agg_series(grouped_df, [["lev2"]])
    expected_levels = ["X", "Y"]
    assert sorted(levels) == sorted(expected_levels)


def test_2lev():
    grouped_df = pandas.DataFrame(
        data={"lev1": ["A", "A", "B", "B"], "lev2": ["X", "Y", "X", "Y"],}
    )

    levels = get_agg_series(grouped_df, [["lev1", "lev2"]])

    expected_levels = ["A_X", "A_Y", "B_X", "B_Y"]

    assert sorted(levels) == sorted(expected_levels)


def test_hierarchichal():
    hier_df = pandas.DataFrame(
        data={"lev1": ["A", "A", "A", "B", "B"], "lev2": ["X", "Y", "Z", "X", "Y"],}
    )

    levels = get_agg_series(hier_df, [["lev1"], ["lev1", "lev2"]])
    expected_levels = ["A", "B", "A_X", "A_Y", "A_Z", "B_X", "B_Y"]
    assert sorted(levels) == sorted(expected_levels)


def test_grouped():
    hier_df = pandas.DataFrame(
        data={"lev1": ["A", "A", "A", "B", "B"], "lev2": ["X", "Y", "Z", "X", "Y"],}
    )

    hierarchy = [["lev1"], ["lev2"], ["lev1", "lev2"]]
    levels = get_agg_series(hier_df, hierarchy)
    expected_levels = ["A", "B", "X", "Y", "Z", "A_X", "A_Y", "A_Z", "B_X", "B_Y"]
    assert sorted(levels) == sorted(expected_levels)


def test_grouped_create_df():
    hier_df = pandas.DataFrame(
        data={
            "ds": ["2020-01", "2020-02"] * 5,
            "lev1": ["A", "A", "A", "A", "A", "A", "B", "B", "B", "B"],
            "lev2": ["X", "X", "Y", "Y", "Z", "Z", "X", "X", "Y", "Y"],
            "val": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        }
    )

    level_names = ["lev1", "lev2"]
    hierarchy = [["lev1"], ["lev2"]]
    gts_df, sum_mat, sum_mat_labels = get_hierarchichal_df(
        hier_df,
        level_names=level_names,
        hierarchy=hierarchy,
        date_colname="ds",
        val_colname="val",
    )

    expected_columns = [
        "A_X",
        "A_Y",
        "A_Z",
        "B_X",
        "B_Y",
        "A",
        "B",
        "X",
        "Y",
        "Z",
        "total",
    ]
    assert sorted(list(gts_df.columns)) == sorted(expected_columns)


def test_parent_child():
    grouped_df = pandas.DataFrame(
        data={"lev1": ["A", "A", "B"], "lev2": ["X", "Y", "Z"],}
    )

    levels = get_agg_series(grouped_df, [["lev1", "lev2"]])
    expected_levels = ["A_X", "A_Y", "B_Z"]
    assert sorted(levels) == sorted(expected_levels)


def test_create_bl_str_col():
    grouped_df = pandas.DataFrame(
        data={"lev1": ["A", "A", "B"], "lev2": ["X", "Y", "Z"],}
    )

    col = _create_bl_str_col(grouped_df, ["lev1", "lev2"])

    assert col == ["A_X", "A_Y", "B_Z"]



import logging
from typing import NamedTuple, Union

import numpy
import pandas
from statsmodels.tsa.holtwinters import ExponentialSmoothing
from statsmodels.tsa.statespace.sarimax import SARIMAX

from hts._t import ModelT, NAryTreeT, TimeSeriesModelT, TransformT
from hts.core.exceptions import InvalidArgumentException
from hts.hierarchy import HierarchyTree
from hts.transforms import BoxCoxTransformer, FunctionTransformer

logger = logging.getLogger(__name__)


class TimeSeriesModel(TimeSeriesModelT):
    """ Base class for the implementation of the underlying models.
        Inherits from scikit-learn base classes
    """

    def __init__(
        self, kind: str, node: HierarchyTree, transform: TransformT = False, **kwargs
    ):
        """
        Parameters
        ----------
        kind : str
            One of `prophet`, `sarimax`, `auto-arima`, `holt-winters`
        node : HierarchyTree
            Node
        transform : Bool or NamedTuple
        kwargs
            Keyword arguments to be passed to the model instantiation. See the documentation
            of each of the actual model implementations for a more comprehensive treatment
        """

        if kind not in ModelT.names():
            raise InvalidArgumentException(
                f'Model {kind} not valid. Pick one of: {" ".join(ModelT.names())}'
            )

        self.kind = kind
        self.node = node
        self.transform_function = self._set_transform(transform=transform)
        self.model = self.create_model(**kwargs)
        self.forecast = None
        self.residual = None
        self.mse = None

    def _set_transform(self, transform: TransformT):
        if transform is False or transform is None:
            return FunctionTransformer(func=self._no_func, inv_func=self._no_func)
        elif transform is True:
            return BoxCoxTransformer()
        elif isinstance(transform, tuple):
            if not hasattr(transform, "func") or not hasattr(transform, "inv_func"):
                raise ValueError(
                    "If passing a NamedTuple, it must have a `func` and `inv_func` parameters"
                )
            return FunctionTransformer(
                func=getattr(transform, "func"), inv_func=getattr(transform, "inv_func")
            )
        else:
            raise ValueError(
                "Invalid transform passed. Use either `True` for default boxcox transform or "
                "a `NamedTuple(func: Callable, inv_func: Callable)` for custom transforms"
            )

    def _set_results_return_self(self, in_sample, y_hat):
        in_sample = self.transform_function.inverse_transform(in_sample)
        y_hat = self.transform_function.inverse_transform(y_hat)
        self.forecast = pandas.DataFrame(
            {"yhat": numpy.concatenate([in_sample, y_hat])}
        )
        self.residual = (in_sample - self._get_transformed_data(as_series=True)).values
        self.mse = numpy.mean(numpy.array(self.residual) ** 2)
        return self

    def _get_transformed_data(
        self, as_series: bool = False
    ) -> Union[pandas.DataFrame, pandas.Series]:
        key = self.node.key
        value = self.node.item
        transformed = self.transform_function.transform(value[key])
        if as_series:
            return pandas.Series(transformed)
        else:
            return pandas.DataFrame({key: transformed})

    def create_model(self, **kwargs):

        if self.kind == ModelT.holt_winters.name:
            data = self._get_transformed_data()
            model = ExponentialSmoothing(endog=data, **kwargs)

        elif self.kind == ModelT.auto_arima.name:
            try:
                from pmdarima import AutoARIMA
            except ImportError:  # pragma: no cover
                logger.error(
                    "pmdarima not installed, so auto_arima won't work. Exiting."
                    "Install it with: pip install scikit-hts[auto_arima]"
                )
                return
            model = AutoARIMA(**kwargs)

        elif self.kind == ModelT.sarimax.name:
            as_df = self.node.item
            end = self._get_transformed_data(as_series=True)
            if self.node.exogenous:
                ex = as_df[self.node.exogenous]
            else:
                ex = None
            model = SARIMAX(endog=end, exog=ex, **kwargs)
        else:
            raise
        return model

    def fit(self, **fit_args) -> "TimeSeriesModel":
        raise NotImplementedError

    def predict(self, node: HierarchyTree, **predict_args):
        raise NotImplementedError

    def fit_predict(self, node: HierarchyTree, **kwargs):
        return self.fit().predict(node)



import logging
import warnings

import pandas
from statsmodels.tools.sm_exceptions import ConvergenceWarning

from hts._t import ModelT
from hts.hierarchy import HierarchyTree
from hts.model.base import TimeSeriesModel


class AutoArimaModel(TimeSeriesModel):
    """
    Wrapper class around ``pmdarima.AutoARIMA``

    Attributes
    ----------
    model : pmdarima.AutoARIMA
        The instance of the model

    mse : float
        MSE for in-sample predictions

    residual : numpy.ndarry
        Residuals for the in-sample predictions

    forecast : pandas.DataFramer
        The forecast for the trained model

    Methods
    -------
    fit(self, **fit_args)
        Fits underlying models to the data, passes kwargs to ``AutoARIMA``

    predict(self, node, steps_ahead: int = 10, alpha: float = 0.05)
        Predicts the n-step ahead forecast. Exogenous variables are required if models were
        fit using them
    """

    def __init__(self, node: HierarchyTree, **kwargs):
        super().__init__(ModelT.auto_arima.name, node, **kwargs)

    def fit(self, **fit_args) -> "TimeSeriesModel":
        as_df = self.node.item
        end = self._get_transformed_data(as_series=True)
        if self.node.exogenous:
            ex = as_df[self.node.exogenous]
        else:
            ex = None
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            warnings.filterwarnings("ignore", category=ConvergenceWarning)
            self.model = self.model.fit(y=end, X=ex, **fit_args)
        return self

    def predict(
        self, node, steps_ahead=10, alpha=0.05, exogenous_df: pandas.DataFrame = None
    ):
        if self.node.exogenous:
            ex = node.item
        else:
            ex = None
        in_sample_preds = self.model.predict_in_sample(X=ex, alpha=alpha)
        y_hat = self.model.predict(X=exogenous_df, alpha=alpha, n_periods=steps_ahead)
        return self._set_results_return_self(in_sample_preds, y_hat)

    def fit_predict(self, node: HierarchyTree, steps_ahead=10, alpha=0.05, **fit_args):
        return self.fit(**fit_args).predict(
            node=node, steps_ahead=steps_ahead, alpha=alpha
        )


class SarimaxModel(TimeSeriesModel):
    """
    Wrapper class around ``statsmodels.tsa.statespace.sarimax.SARIMAX``

    Attributes
    ----------
    model : SARIMAX
        The instance of the model

    mse : float
        MSE for in-sample predictions

    residual : numpy.ndarry
        Residuals for the in-sample predictions

    forecast : pandas.DataFramer
        The forecast for the trained model

    Methods
    -------
    fit(self, **fit_args)
        Fits underlying models to the data, passes kwargs to ``SARIMAX``

    predict(self, node, steps_ahead: int = 10, alpha: float = 0.05)
        Predicts the n-step ahead forecast. Exogenous variables are required if models were
        fit using them
    """

    def __init__(self, node: HierarchyTree, **kwargs):
        super().__init__(ModelT.sarimax.name, node, **kwargs)

    def fit(self, **fit_args) -> "TimeSeriesModel":
        self.model = self.model.fit(disp=0, **fit_args)
        return self

    def predict(self, node, steps_ahead=10, alpha=0.05):
        if self.node.exogenous:
            ex = node.item
        else:
            ex = None
        y_hat = self.model.forecast(steps=steps_ahead, exog=ex).values
        in_sample_preds = self.model.get_prediction(
            dynamic=False, exog=ex
        ).predicted_mean
        return self._set_results_return_self(in_sample_preds, y_hat)

    def fit_predict(self, node: HierarchyTree, steps_ahead=10, alpha=0.05, **fit_args):
        return self.fit(**fit_args).predict(
            node=node, steps_ahead=steps_ahead, alpha=alpha
        )



from collections import namedtuple

import numpy
import pandas
import pytest

from hts.transforms import BoxCoxTransformer, FunctionTransformer


def test_fit_transform(hierarchical_mv_data):
    tf = BoxCoxTransformer()
    df = hierarchical_mv_data
    transformed = tf.fit_transform(df["WF-01"])
    assert isinstance(transformed, numpy.ndarray)
    assert tf.lam is not None

    transformed = tf.fit_transform(df["WF-01"].replace(0, 6))
    assert isinstance(transformed, numpy.ndarray)
    assert tf.lam is not None

    with pytest.raises(ValueError):
        _ = tf.fit_transform(df["precipitation"].replace(0, -10))


def test_inv_transform(hierarchical_mv_data):
    tf = BoxCoxTransformer()
    df = hierarchical_mv_data

    base = df["WF-01"].replace(0, 6)
    transformed = tf.fit_transform(base)

    inv_transformed = tf.inverse_transform(transformed)
    assert numpy.allclose(inv_transformed, base, atol=0.1)

    inv_transformed = tf.inverse_transform(pandas.Series(transformed))
    assert numpy.allclose(inv_transformed, base, atol=0.1)


def test_custom_transform(hierarchical_mv_data):
    def scale(val, src, dst):
        return ((val - src[0]) / (src[1] - src[0])) * (dst[1] - dst[0]) + dst[0]

    df = hierarchical_mv_data
    values = df["WF-01"]
    values_centered = scale(values, [min(values), max(values)], [0, 1])
    Transform = namedtuple("Transform", ["func", "inv_func"])

    transform_pos_neg = Transform(func=numpy.exp, inv_func=lambda x: -x)
    transform_isomorphic = Transform(func=numpy.exp, inv_func=numpy.log)

    function_transform_pos_neg = FunctionTransformer(
        func=getattr(transform_pos_neg, "func"),
        inv_func=getattr(transform_pos_neg, "inv_func"),
    )
    function_transform_isomorphic = FunctionTransformer(
        func=getattr(transform_isomorphic, "func"),
        inv_func=getattr(transform_isomorphic, "inv_func"),
    )

    transformed = function_transform_pos_neg.fit_transform(values_centered)
    assert isinstance(transformed, numpy.ndarray)
    assert not (transformed < 0).any()

    inv_transformed = function_transform_pos_neg.inverse_transform(values_centered)
    assert isinstance(inv_transformed, pandas.Series)
    assert not (inv_transformed.values > 0).any()

    transformed = function_transform_isomorphic.fit_transform(values_centered)
    inv_transformed = function_transform_isomorphic.inverse_transform(
        pandas.Series(transformed)
    )
    assert numpy.allclose(inv_transformed, values_centered, atol=0.1)



import numpy

from hts._t import MethodT
from hts.core.exceptions import InvalidArgumentException
from hts.functions import (
    forecast_proportions,
    optimal_combination,
    proportions,
    y_hat_matrix,
)
from hts.hierarchy.utils import make_iterable


class RevisionMethod(object):
    def __init__(
        self, name: str, sum_mat: numpy.ndarray, transformer,
    ):
        self.name = name
        self.transformer = transformer
        self.sum_mat = sum_mat

    def _new_mat(self, y_hat_mat) -> numpy.ndarray:
        new_mat = numpy.empty([y_hat_mat.shape[0], self.sum_mat.shape[0]])
        for i in range(y_hat_mat.shape[0]):
            new_mat[i, :] = numpy.dot(self.sum_mat, numpy.transpose(y_hat_mat[i, :]))
        return new_mat

    def _y_hat_matrix(self, forecasts) -> numpy.ndarray:
        n_cols = len(list(forecasts.keys())) + 1
        if self.name == "BU":
            keys = list(forecasts.keys())[
                n_cols - self.sum_mat.shape[1] - 1 : n_cols - 1
            ]
        else:
            keys = range(n_cols - self.sum_mat.shape[1] - 1, n_cols - 1)
        return y_hat_matrix(forecasts, keys=keys)

    def revise(self, forecasts=None, mse=None, nodes=None) -> numpy.ndarray:
        """


        Parameters
        ----------
        forecasts
        mse
        nodes

        Returns
        -------

        """
        if self.name == MethodT.NONE.name:
            return y_hat_matrix(forecasts=forecasts)

        if self.name in [MethodT.OLS.name, MethodT.WLSS.name, MethodT.WLSV.name]:
            return optimal_combination(
                forecasts=forecasts, sum_mat=self.sum_mat, method=self.name, mse=mse
            )

        elif self.name == MethodT.BU.name:
            y_hat = self._y_hat_matrix(forecasts)
            return self._new_mat(y_hat)

        elif self.name in [MethodT.AHP.name, MethodT.PHA.name]:
            if self.transformer:
                for node in make_iterable(nodes, prop=None):
                    node.item[node.key] = self.transformer.inverse_transform(
                        node.item[node.key]
                    )
            y_hat = proportions(
                nodes=nodes, forecasts=forecasts, sum_mat=self.sum_mat, method=self.name
            )
            return self._new_mat(y_hat)

        elif self.name == MethodT.FP.name:
            return forecast_proportions(forecasts, nodes)

        else:
            raise InvalidArgumentException("Revision model name is invalid")



import os
from multiprocessing import cpu_count

from hts._t import MethodT, ModelT

n_cores = int(os.getenv("NUMBER_OF_CPUS") or cpu_count())

MODEL = ModelT.prophet.value
REVISION = MethodT.OLS.value
LOW_MEMORY = False
CHUNKSIZE = None
N_PROCESSES = max(1, n_cores // 2)
PROFILING = False
DISABLE_PROGRESSBAR = False
SHOW_WARNINGS = False
PARALLELISATION = True



"""
This module contains the Distributor class, such objects are used to distribute the calculation of features.
Essentially, a Distributor organizes the application of feature calculators to data chunks.
Design of this module by Nils Braun
"""

import itertools
import math
import warnings
from collections import Iterable
from functools import partial
from multiprocessing import Pool

from tqdm import tqdm


def _function_with_partly_reduce(chunk_list, map_function, kwargs):
    """
    Small helper function to call a function (map_function)
    on a list of data chunks (chunk_list) and convert the results into
    a flattened list.
    This function is used to send chunks of data with a size larger than 1 to
    the workers in parallel and process these on the worker.

    Parameters
    ----------
    chunk_list : List
         A list of data chunks to process.
    map_function : Callable
        Function to be partially applied
    kwargs : dict
        Keyword arguments to be passed to the map_function
    Returns
    -------
    List
        A list of the results of the function evaluated on each chunk and flattened.
    """

    kwargs = kwargs or {}
    results = (map_function(chunk, kwargs) for chunk in chunk_list)
    return list(results)


def initialize_warnings_in_workers(show_warnings):  # pragma: no cover
    """
    Small helper function to initialize warnings module in multiprocessing workers.
    On Windows, Python spawns fresh processes which do not inherit from warnings
    state, so warnings must be enabled/disabled before running computations.
    """
    warnings.catch_warnings()
    if not show_warnings:
        warnings.simplefilter("ignore")
    else:
        warnings.simplefilter("default")


class DistributorBaseClass:
    """
    The distributor abstract base class.
    The main purpose of the instances of the DistributorBaseClass subclasses is to evaluate a function
    (called map_function) on a list of data items (called data).
    This is done on chunks of the data, meaning, that the DistributorBaseClass classes will chunk the data into chunks,
    distribute the data and apply the feature calculator functions from
    Dependent on the implementation of the distribute function, this is done in parallel or using a cluster of nodes.
    """

    @staticmethod
    def partition(data, chunk_size):
        """
        This generator chunks a list of data into slices of length chunk_size. If the chunk_size is not a divider of the
        data length, the last slice will be shorter than chunk_size.

        Parameters
        ----------
        data : List
            The data to chunk
        chunk_size : int
            Each chunks size. The last chunk may be smaller.
        Returns
        -------
        Generator
            A generator producing the chunks of data.
        """

        iterable = iter(data)
        while True:
            next_chunk = list(itertools.islice(iterable, chunk_size))
            if not next_chunk:
                return

            yield next_chunk

    def __init__(self):
        """
        Constructs the DistributorBaseClass class
        """
        self.n_workers = None
        self.disable_progressbar = False
        self.progressbar_title = None

    def calculate_best_chunk_size(self, data_length):
        """
        Calculates the best chunk size for a list of length data_length. The current implemented formula is more or
        less an empirical result for multiprocessing case on one machine.

        Parameters
        ----------
        data_length : int
             A length which defines how many calculations there need to be.
        Returns
        -------
        int
            The calculated chunk size

        """
        chunk_size, extra = divmod(data_length, self.n_workers * 5)
        if extra:
            chunk_size += 1
        return chunk_size

    def map_reduce(
        self,
        map_function,
        data,
        function_kwargs=None,
        chunk_size=None,
        data_length=None,
    ):
        """
        This method contains the core functionality of the DistributorBaseClass class.
        It maps the map_function to each element of the data and reduces the results to return a flattened list.
        How the jobs are calculated, is determined by the class'
        :func:`hts.utilities.distribution.DistributorBaseClass.distribute` method,
        which can distribute the jobs in multiple threads, across multiple processing units etc.
        To not transport each element of the data individually, the data is split into chunks, according to the chunk
        size (or an empirical guess if none is given). By this, worker processes not tiny but adequate sized parts of
        the data.

        Parameters
        ----------
        map_function : Callable
            Function to apply to each data item.
        data : List
            The data to use in the calculation
        function_kwargs : Dict
            Parameters for the map function
        chunk_size : int
            If given, chunk the data according to this size. If not given, use an empirical value.
        data_length : int
            If the data is a generator, you have to set the length here. If it is none, the
            length is deduced from the len of the data.

        Returns
        -------
        List
            The calculated results
        """
        if data_length is None:
            data_length = len(data)

        if not chunk_size:
            chunk_size = self.calculate_best_chunk_size(data_length)

        chunk_generator = self.partition(data, chunk_size=chunk_size)

        map_kwargs = {"map_function": map_function, "kwargs": function_kwargs}

        total_number_of_expected_results = math.ceil(data_length / chunk_size)
        result = tqdm(
            self.distribute(_function_with_partly_reduce, chunk_generator, map_kwargs),
            total=total_number_of_expected_results,
            desc=self.progressbar_title,
            disable=self.disable_progressbar,
        )

        result = list(itertools.chain.from_iterable(result))

        return result

    def distribute(self, func, partitioned_chunks, kwargs):
        """
        This abstract base function distributes the work among workers, which can be threads or nodes in a cluster.
        Must be implemented in the derived classes.

        Parameters
        ----------
        func : Callable
            Function to send to each worker.
        partitioned_chunks : List
            List of data chunks, each chunk is processed by one woker
        kwargs : Dict
            Parameters for the map function
        Raises
        -------
        NotImplementedError
        """
        raise NotImplementedError

    def close(self):
        """
        Abstract base function to clean the DistributorBaseClass after use, e.g. close the connection to a DaskScheduler
        """
        pass


class MapDistributor(DistributorBaseClass):
    """
    Distributor using the python build-in map, which calculates each job sequentially one after the other.
    """

    def __init__(self, disable_progressbar=False, progressbar_title="Fitting Models"):
        """
        Creates a new MapDistributor instance

        Parameters
        ----------
        disable_progressbar : bool
            Disables tqdm's progressbar
        progressbar_title : str
            Title of progressbar
        """

        super().__init__()
        self.disable_progressbar = disable_progressbar
        self.progressbar_title = progressbar_title

    def distribute(self, func, partitioned_chunks, kwargs):
        """
        Calculates the features in a sequential fashion by pythons map command

        Parameters
        ----------
        func : Callable
            Function to send to each worker.
        partitioned_chunks : List
            List of data chunks, each chunk is processed by one woker
        kwargs : Dict
            Parameters for the map function
        Returns
        -------
        List
            The result of the calculation as a list - each item should be the result of the application of func
            to a single element.

        """
        return map(partial(func, **kwargs), partitioned_chunks)

    def calculate_best_chunk_size(self, data_length):
        """
        For the map command, which calculates the features sequentially, a the chunk_size of 1 will be used.

        Parameters
        ----------
        data_length : int
            1
        """
        return 1


class LocalDaskDistributor(DistributorBaseClass):
    """
    Distributor using a local dask cluster and inproc communication.
    """

    def __init__(self, n_workers):
        """
        Initiates a LocalDaskDistributor instance.

        Parameters
        ----------
        n_workers : int
            How many workers should the local dask cluster have?
        """

        super().__init__()
        import tempfile

        from distributed import Client, LocalCluster

        # attribute .local_dir_ is the path where the local dask workers store temporary files
        self.local_dir_ = tempfile.mkdtemp()
        cluster = LocalCluster(
            n_workers=n_workers, processes=False, local_dir=self.local_dir_
        )

        self.client = Client(cluster)
        self.n_workers = n_workers

    def distribute(self, func, partitioned_chunks, kwargs):
        """
        Calculates the features in a parallel fashion by distributing the map command to the dask workers on a local
        machine

        Parameters
        ----------
        func : Callable
            Function to send to each worker.
        partitioned_chunks : List
            List of data chunks, each chunk is processed by one woker
        kwargs : Dict
            Parameters for the map function
        Returns
        -------
        List
            The result of the calculation as a list - each item should be the result of the application of func
            to a single element.
        """

        if isinstance(partitioned_chunks, Iterable):
            # since dask 2.0.0 client map no longer accepts iterables
            partitioned_chunks = list(partitioned_chunks)
        result = self.client.gather(
            self.client.map(partial(func, **kwargs), partitioned_chunks)
        )
        return result

    def close(self):
        """
        Closes the connection to the local Dask Scheduler
        """
        self.client.close()


class ClusterDaskDistributor(DistributorBaseClass):
    """
    Distributor using a dask cluster, meaning that the calculation is spread over a cluster
    """

    def __init__(self, address):
        """
        Sets up a distributor that connects to a Dask Scheduler to distribute the calculaton of the features

        Parameters
        ----------
        address : str
            The ip address and port number of the Dask Scheduler
        """

        super().__init__()
        from distributed import Client

        self.client = Client(address=address)

    def calculate_best_chunk_size(self, data_length):
        """
        Uses the number of dask workers in the cluster (during execution time, meaning when you start the extraction)
        to find the optimal chunk_size.

        Parameters
        ----------
        data_length: int
            A length which defines how many calculations there need to be.
        """

        n_workers = len(self.client.scheduler_info()["workers"])
        chunk_size, extra = divmod(data_length, n_workers * 5)
        if extra:
            chunk_size += 1
        return chunk_size

    def distribute(self, func, partitioned_chunks, kwargs):
        """
        Calculates the features in a parallel fashion by distributing the map command to the dask workers on a cluster

        Parameters
        ----------
        func : Callable
            Function to send to each worker.
        partitioned_chunks : List
            List of data chunks, each chunk is processed by one woker
        kwargs : Dict
            Parameters for the map function
        Returns
        -------
        List
            The result of the calculation as a list - each item should be the result of the application of func
            to a single element
        """

        if isinstance(partitioned_chunks, Iterable):
            # since dask 2.0.0 client map no longer accepts iterables
            partitioned_chunks = list(partitioned_chunks)
        result = self.client.gather(
            self.client.map(partial(func, **kwargs), partitioned_chunks)
        )
        return result

    def close(self):
        """
        Closes the connection to the Dask Scheduler
        """
        self.client.close()


class MultiprocessingDistributor(DistributorBaseClass):
    """
    Distributor using a multiprocessing Pool to calculate the jobs in parallel on the local machine.
    """

    def __init__(
        self,
        n_workers,
        disable_progressbar=False,
        progressbar_title="Feature Extraction",
        show_warnings=True,
    ):
        """
        Creates a new MultiprocessingDistributor instance

        Parameters
        ----------
        n_workers : int
            How many workers should the multiprocessing pool have?
        disable_progressbar : bool
            Disables tqdm's progressbar
        progressbar_title : str
            Title of progressbar
        """

        super().__init__()
        self.pool = Pool(
            processes=n_workers,
            initializer=initialize_warnings_in_workers,
            initargs=(show_warnings,),
        )
        self.n_workers = n_workers
        self.disable_progressbar = disable_progressbar
        self.progressbar_title = progressbar_title

    def distribute(self, func, partitioned_chunks, kwargs):
        """
        Calculates the features in a parallel fashion by distributing the map command to a thread pool

        Parameters
        ----------
        func : Callable
            Function to send to each worker.
        partitioned_chunks : List
            List of data chunks, each chunk is processed by one woker
        kwargs : Dict
            Parameters for the map function
        Returns
        -------
        List
            The result of the calculation as a list - each item should be the result of the application of func
            to a single element.
        """

        return self.pool.imap_unordered(partial(func, **kwargs), partitioned_chunks)

    def close(self):
        """
        Collects the result from the workers and closes the thread pool.
        """
        self.pool.close()
        self.pool.terminate()
        self.pool.join()



import pandas

from hts._t import NAryTreeT
from hts.hierarchy import HierarchyTree


def test_level_order_traversal(n_tree):
    assert n_tree.level_order_traversal() == [[3], [2, 2, 2], [2, 2, 2, 2, 2, 2]]
    bottom = n_tree.get_node("aab")
    for i in bottom.level_order_traversal():
        assert i is None


def test_get_heights(n_tree):
    assert [sum(i) for i in n_tree.level_order_traversal()] == [3, 6, 12]


def test_get_node(n_tree):
    bab = n_tree.get_node("bab")
    assert isinstance(bab, NAryTreeT)
    assert bab.item == 6
    assert bab.is_leaf()
    assert n_tree.is_leaf() is False
    assert n_tree.leaf_sum() == 12
    assert n_tree.get_node("doesnotexist") is None


def test_create_from_events(events):
    ht = HierarchyTree.from_geo_events(
        df=events,
        lat_col="start_latitude",
        lon_col="start_longitude",
        nodes=("city", "hex_index_6", "hex_index_7", "hex_index_8"),
        levels=(6, 8),
        resample_freq="1H",
        min_count=0.5,
    )
    assert isinstance(ht, NAryTreeT)
    assert len(ht.children) == events["city"].nunique()


def test_num_nodes(n_tree):
    assert n_tree.num_nodes() == 21


def test_height(n_tree):
    assert n_tree.value_at_height(1) == [3]
    assert n_tree.value_at_height(2) == [2, 2, 2]
    assert n_tree.value_at_height(3) == [2, 2, 2, 2, 2, 2]
    assert n_tree.sum_at_height(3) == 12
    assert n_tree.get_node_height("bb") == 1
    assert n_tree.get_node_height("bbb") == 0
    assert n_tree.get_node_height("a") == 2
    assert n_tree.get_node_height("z") == -1


def test_to_pandas(events):
    ht = HierarchyTree.from_geo_events(
        df=events,
        lat_col="start_latitude",
        lon_col="start_longitude",
        nodes=("city", "hex_index_6", "hex_index_7", "hex_index_8"),
        levels=(6, 8),
        resample_freq="1H",
        min_count=0.5,
    )
    assert isinstance(ht.to_pandas(), pandas.DataFrame)


def test_from_geo_events(events):
    ht = HierarchyTree.from_geo_events(
        df=events,
        lat_col="start_latitude",
        lon_col="start_longitude",
        nodes=("city", "hex_index_6", "hex_index_7", "hex_index_8"),
        levels=(6, 8),
        resample_freq="1H",
        min_count=0.5,
        fillna=True,
    )
    assert isinstance(ht.to_pandas(), pandas.DataFrame)


def test_create_hierarchical_sine_data_tree(hierarchical_sine_data):
    hier = {
        "total": ["a", "b", "c"],
        "a": ["a_x", "a_y"],
        "b": ["b_x", "b_y"],
        "c": ["c_x", "c_y"],
        "a_x": ["a_x_1", "a_x_2"],
        "a_y": ["a_y_1", "a_y_2"],
        "b_x": ["b_x_1", "b_x_2"],
        "b_y": ["b_y_1", "b_y_2"],
        "c_x": ["c_x_1", "c_x_2"],
        "c_y": ["c_y_1", "c_y_2"],
    }
    ht = HierarchyTree.from_nodes(hier, hierarchical_sine_data)
    assert isinstance(ht.to_pandas(), pandas.DataFrame)
    assert ht.key == "total"
    assert len(ht.children) == 3
    for c in ht.children:
        if c.key == "a" or c.key == "b" or c.key == "c":
            assert len(c.children) == 2
        if (
            c.key == "a_x"
            or c.key == "b_x"
            or c.key == "c_x"
            or c.key == "a_y"
            or c.key == "b_y"
            or c.key == "c_y"
        ):
            assert len(c.children) == 4


def test_create_mv_tree(hierarchical_mv_data):

    hier = {
        "total": ["CH", "SLU", "BT", "OTHER"],
        "CH": ["CH-07", "CH-02", "CH-08", "CH-05", "CH-01"],
        "SLU": ["SLU-15", "SLU-01", "SLU-19", "SLU-07", "SLU-02"],
        "BT": ["BT-01", "BT-03"],
        "OTHER": ["WF-01", "CBD-13"],
    }
    exogenous = {
        k: ["precipitation", "temp"]
        for k in hierarchical_mv_data.columns
        if k not in ["precipitation", "temp"]
    }

    ht = HierarchyTree.from_nodes(hier, hierarchical_mv_data, exogenous=exogenous)
    assert isinstance(ht.to_pandas(), pandas.DataFrame)
    assert ht.key == "total"
    assert len(ht.children) == 4
    assert ht.get_node_height("CH") == 1
    assert ht.get_node_height("BT-03") == 0
    assert ht.get_node_height("CBD-13") == 0
    assert ht.get_node_height("SLU") == 1



import numpy
import pandas
import pytest

from hts import HTSRegressor
from hts.convenience import (
    _calculate_errors,
    _sanitize_errors_dict,
    _sanitize_forecasts_dict,
    revise_forecasts,
)


def test_convenience_error_handling(load_df_and_hier_uv):
    _, _ = load_df_and_hier_uv
    dummy = numpy.array([1, 2, 3, 5])
    with pytest.raises(ValueError):
        revise_forecasts("FP", forecasts={"a": dummy}, nodes=None)

    with pytest.raises(ValueError):
        revise_forecasts("OLS", summing_matrix=dummy, forecasts={"a": dummy})


def test_errors_and_residuals_dicts():
    res_a, res_b = [1, 2, 1, 3, 1], [0, 0, 1, 0, 0]
    error_a, error_b = 0.1, 0.999
    errors = {"a": error_a, "b": error_b}
    residuals = {"a": numpy.array(res_a), "b": numpy.array(res_b)}
    assert _sanitize_errors_dict(errors) == errors
    assert _calculate_errors(method="A", errors=errors) == errors
    calculated_errors_from_res = _calculate_errors(method="A", residuals=residuals)
    assert isinstance(calculated_errors_from_res, dict)
    assert (
        _calculate_errors(method="A", errors=errors, residuals=residuals)
        == calculated_errors_from_res
    )
    assert calculated_errors_from_res["a"] == numpy.mean(numpy.array(res_a) ** 2)
    assert calculated_errors_from_res["b"] == numpy.mean(numpy.array(res_b) ** 2)

    with pytest.raises(ValueError):
        _sanitize_errors_dict(errors={"a": "12340"})  # pragma: nocover


def test_sanitize_forecasts():
    values = [[1, 2, 1, 3, 1], [0, 0, 0, 0, 0]]
    forecasts = {"a": numpy.array(values[0]), "b": numpy.array(values[1])}
    sanitized = _sanitize_forecasts_dict(forecasts)
    assert isinstance(sanitized, dict)
    assert isinstance(sanitized["a"].yhat, pandas.Series)

    forecasts = {"a": pandas.Series(values[0]), "b": pandas.Series(values[1])}
    sanitized = _sanitize_forecasts_dict(forecasts)
    assert isinstance(sanitized, dict)
    assert isinstance(sanitized["a"].yhat, pandas.Series)

    forecasts = {"a": pandas.DataFrame({"1": values[0]})}
    sanitized = _sanitize_forecasts_dict(forecasts)
    assert isinstance(sanitized, dict)
    assert isinstance(sanitized["a"].yhat, pandas.Series)

    forecasts = {"a": numpy.array(values), "b": numpy.array(values)}
    with pytest.raises(ValueError):
        _sanitize_forecasts_dict(forecasts)

    forecasts = {"a": pandas.DataFrame({"1": values[0], "2": values[1]})}
    with pytest.raises(ValueError):
        _sanitize_forecasts_dict(forecasts)


def test_convenience(load_df_and_hier_uv):
    hd, hier = load_df_and_hier_uv
    hd = hd.head(200)
    ht = HTSRegressor(model="holt_winters", revision_method="OLS")
    ht = ht.fit(df=hd, nodes=hier)
    preds = ht.predict(steps_ahead=10)

    rev = revise_forecasts(
        "OLS",
        nodes=ht.nodes,
        forecasts=ht.hts_result.forecasts,
        errors=ht.hts_result.errors,
        residuals=ht.hts_result.residuals,
    )
    assert isinstance(rev, pandas.DataFrame)
    assert numpy.allclose(preds.values, rev.values)

    rev_s = revise_forecasts(
        "OLS",
        summing_matrix=ht.sum_mat,
        forecasts=ht.hts_result.forecasts,
        errors=ht.hts_result.errors,
        residuals=ht.hts_result.residuals,
    )
    assert isinstance(rev_s, pandas.DataFrame)
    assert numpy.allclose(rev.values, rev_s.values)

    rev = revise_forecasts("AHP", nodes=ht.nodes, forecasts=ht.hts_result.forecasts)
    assert isinstance(rev, pandas.DataFrame)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path

from setuptools import find_packages, setup

with open("README.rst", encoding="utf-8") as readme_file:
    readme = readme_file.read()


with open("HISTORY.rst") as history_file:
    history = history_file.read()

EXTENSIONS = {"auto_arima", "prophet", "geo", "test", "dev", "distributed", "all"}


def strip_comments(l):
    return l.split("#", 1)[0].strip()


def _pip_requirement(req, *root):
    if req.startswith("-r "):
        _, path = req.split()
        return reqs(*root, *path.split("/"))
    return [req]


def _reqs(*f):
    path = (Path.cwd() / "reqs").joinpath(*f)
    with path.open() as fh:
        reqs = [strip_comments(l) for l in fh.readlines()]
        return [_pip_requirement(r, *f[:-1]) for r in reqs if r]


def reqs(*f):
    return [req for subreq in _reqs(*f) for req in subreq]


def extras(*p):
    """Parse requirement in the requirements/extras/ directory."""
    return reqs("extras", *p)


def extras_require():
    """Get map of all extra requirements."""
    return {x: extras(x + ".txt") for x in EXTENSIONS}


install_requires = reqs("base.txt")
test_requires = (
    extras("prophet.txt")
    + extras("auto_arima.txt")
    + extras("distributed.txt")
    + extras("test.txt")
    + install_requires
)

setup(
    author="Carlo Mazzaferro",
    author_email="carlo.mazzaferro@gmail.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Hierarchical Time Series forecasting",
    install_requires=install_requires,
    extras_require=extras_require(),
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="scikit-hts",
    name="scikit-hts",
    packages=find_packages(include=["hts"]),
    test_suite="tests",
    tests_require=test_requires,
    version="0.5.10",
    zip_safe=False,
)



from random import choice
from typing import Dict, List, Tuple

import numpy as np
import pandas

from hts._t import MethodT, NAryTreeT
from hts.hierarchy import make_iterable


def to_sum_mat(
    ntree: NAryTreeT = None, node_labels: List[str] = None
) -> Tuple[np.ndarray, List[str]]:
    """
    This function creates a summing matrix for the bottom up and optimal combination approaches
    All the inputs are the same as above
    The output is a summing matrix, see Rob Hyndman's "Forecasting: principles and practice" Section 9.4

    Parameters
    ----------
    ntree : NAryTreeT

    node_labels : List[str]
        Labels corresponing to node names/ summing matrix. Get from hts.functions.get_hierarchichal_df(...)

    Returns
    -------
    numpy.ndarray
        Summing matrix.

    List[str]
        Row order list of the level in the hierarchy represented by each row in the summing matrix.

    """
    if node_labels:
        columns = len(node_labels[-1])
    elif ntree:
        nodes = ntree.level_order_traversal()
        node_labels = ntree.get_level_order_labels()
        num_at_level = list(map(sum, nodes))
        columns = num_at_level[-1]
    else:
        raise ValueError(
            "Must pass either ntree or node_labels to the function. Neither was received."
        )

    # Initialize summing matrix with bottom level rows
    sum_mat = np.identity(columns)

    # Names of each row in summing matrix.
    sum_mat_labels = []

    # Bottom level matrix labels, with indices correspoding to column in summing matrix
    bl_mat_idx_ref = node_labels[-1]

    # Skip total and bottom level of tree. Rows added outside of loop.
    for level in node_labels[1:-1]:
        for label in level:
            # Exclude duplicates specified in tree
            if label not in sum_mat_labels:
                row = []
                for bl_element in bl_mat_idx_ref:
                    # Check if the bottom level element is part of label
                    is_component = all(
                        [True if x in bl_element else False for x in label.split("_")]
                    )
                    if is_component:
                        row.append(1)
                    else:
                        row.append(0)

                # Add row correspoding to label to top of summing matrix
                row = np.array(row)
                sum_mat = np.vstack((row, sum_mat))
                sum_mat_labels.append(label)

    # Add top as first row in summing matrix
    top = np.ones(columns)
    sum_mat = np.vstack((top, sum_mat))

    # Reverse list of labels to match summing matrix, since vstack and append worked in the opposite order.
    # Not currently returned, but could be for information or matrix alignment.
    sum_mat_labels.reverse()
    sum_mat_labels = ["total"] + sum_mat_labels + bl_mat_idx_ref

    return sum_mat, sum_mat_labels


def project(
    hat_mat: np.ndarray, sum_mat: np.ndarray, optimal_mat: np.ndarray
) -> np.ndarray:
    new_mat = np.empty([hat_mat.shape[0], sum_mat.shape[0]])
    for i in range(hat_mat.shape[0]):
        new_mat[i, :] = np.dot(optimal_mat, np.transpose(hat_mat[i, :]))
    return new_mat


def y_hat_matrix(forecasts, keys=None):
    if not keys:
        keys = forecasts.keys()
    first = list(forecasts.keys())[0]
    y_hat_mat = np.zeros([len(forecasts[first].yhat), 1])
    for key in keys:
        f1 = np.array(forecasts[key].yhat)
        f2 = f1[:, np.newaxis]
        if np.all(y_hat_mat == 0):
            y_hat_mat = f2
        else:
            y_hat_mat = np.concatenate((y_hat_mat, f2), axis=1)
    return y_hat_mat


def optimal_combination(
    forecasts: Dict[str, pandas.DataFrame],
    sum_mat: np.ndarray,
    method: str,
    mse: Dict[str, float],
):
    """
    Produces the optimal combination of forecasts by trace minimization (as described by
    Wickramasuriya, Athanasopoulos, Hyndman in "Optimal Forecast Reconciliation for Hierarchical and Grouped Time
    Series Through Trace Minimization")

    Parameters
    ----------
    forecasts : dict
        Dictionary of pandas.DataFrames containing the future predictions
    sum_mat : np.ndarray
        The summing  matrix
    method : str
        One of:
            - OLS (ordinary least squares)
            - WLSS (structurally weighted least squares)
            - WLSV (variance weighted least squares)
    mse

    Returns
    -------

    """
    hat_mat = y_hat_matrix(forecasts)
    transpose = np.transpose(sum_mat)

    if method == MethodT.OLS.name:
        ols = np.dot(
            np.dot(sum_mat, np.linalg.inv(np.dot(transpose, sum_mat))), transpose
        )
        return project(hat_mat=hat_mat, sum_mat=sum_mat, optimal_mat=ols)
    elif method == MethodT.WLSS.name:
        diag = np.diag(np.transpose(np.sum(sum_mat, axis=1)))
    elif method == MethodT.WLSV.name:
        diag = [mse[key] for key in mse.keys()]
        diag = np.diag(np.flip(np.hstack(diag) + 0.0000001, 0))
    else:
        raise ValueError("Invalid method")

    # S*inv(S'S)*S'
    optimal_mat = np.dot(
        np.dot(
            np.dot(
                sum_mat,
                np.linalg.inv(np.dot(np.dot(transpose, np.linalg.inv(diag)), sum_mat)),
            ),
            transpose,
        ),
        np.linalg.inv(diag),
    )

    return project(hat_mat=hat_mat, sum_mat=sum_mat, optimal_mat=optimal_mat)


def proportions(nodes, forecasts, sum_mat, method=MethodT.PHA.name):
    n_cols = len(list(forecasts.keys()))
    fcst = forecasts[list(forecasts.keys())[0]].yhat
    fcst = fcst[:, np.newaxis]
    num_bts = sum_mat.shape[1]

    cols = [n.key for n in [nodes] + nodes.traversal_level()][
        (n_cols - num_bts) : n_cols
    ]

    bts_dat = nodes.to_pandas()[cols]
    if method == MethodT.AHP.name:
        divs = np.divide(np.transpose(np.array(bts_dat)), np.array(nodes.get_series()))
        props = divs.mean(1)
        props = props[:, np.newaxis]
    elif method == MethodT.PHA.name:
        bts_sum = bts_dat.sum(0)
        top_sum = sum(nodes.get_series())
        props = bts_sum / top_sum
        props = props[:, np.newaxis]
    else:
        raise ValueError("Invalid method")
    return np.dot(np.array(fcst), np.transpose(props))


def forecast_proportions(forecasts, nodes):
    """
    Cons:
       Produces biased revised forecasts even if base forecasts are unbiased
    """
    n_cols = len(list(forecasts.keys())) + 1

    levels = nodes.get_height()
    column = 0
    first_node = 1

    key = choice(list(forecasts.keys()))
    new_mat = np.empty([len(forecasts[key].yhat), n_cols - 1])
    new_mat[:, 0] = forecasts[key].yhat

    as_iterable = make_iterable(nodes, prop=None)

    for level in range(levels - 1):
        for i, node in enumerate(nodes.level_order_traversal()[level]):
            num_child = node
            last_node = first_node + num_child
            base_fcst = np.array(
                [forecasts[k.key].yhat[:] for k in as_iterable[first_node:last_node]]
            )
            print(base_fcst.shape)
            fore_sum = np.sum(base_fcst, axis=0)
            fore_sum = fore_sum[:, np.newaxis]
            if column == 0:
                rev_top = np.array(forecasts["total"].yhat)
                rev_top = rev_top[:, np.newaxis]
            else:
                rev_top = np.array(new_mat[:, column])
                rev_top = rev_top[:, np.newaxis]
            new_mat[:, first_node:last_node] = np.divide(
                np.multiply(np.transpose(base_fcst), rev_top), fore_sum
            )
            column += 1
            first_node += num_child
    return new_mat


def get_agg_series(df: pandas.DataFrame, levels: List[List[str]]) -> List[str]:
    """
    Get aggregate level series names.

    Parameters
    ----------
    df : pandas.DataFrame
        Tabular data.
    levels : List[List[str]]
        List of lists containing the desired level of aggregation.

    Returns
    -------
    List[str]
        Aggregate series names.
    """
    grouped_levels = []
    for level in levels:
        cross_vals = list(
            "_".join(x for x in y) for y in df[level].drop_duplicates().values
        )
        grouped_levels += cross_vals
    return grouped_levels


def _create_bl_str_col(df: pandas.DataFrame, level_names: List[str]) -> List[str]:
    """
    Concatenate the column values of all the specified level_names by row into a single column.

    Parameters
    ----------
    df : pandas.DataFrame
        Tabular data.
    level_names : List[str]
        Levels in the hierarchy.

    Returns
    -------
    List[str]
        Concatendated column values by row.
    """
    return list("_".join(x for x in y) for y in df[level_names].values)


def add_agg_series_to_df(
    df: pandas.DataFrame, grouped_levels: List[str], bottom_levels: List[str]
) -> pandas.DataFrame:
    """
    Add aggregate series columns to wide dataframe.

    Parameters
    ----------
    df : pandas.DataFrame
        Wide dataframe containing bottom level series.
    grouped_levels : List[str]
        Grouped level, underscore delimited, column names.
    bottom_levels : List[str]
        Bottom level, underscore delimited, column names.

    Returns
    -------
    pandas.DataFrame
        Wide dataframe with all series in hierarchy.
    """
    component_cols = _get_bl(grouped_levels, bottom_levels)
    # Add series as specified grouping levels
    for i, cols in enumerate(component_cols):
        df[grouped_levels[i]] = df[cols].sum(axis=1)
    return df


def _get_bl(grouped_levels: List[str], bottom_levels: List[str]) -> List[List[str]]:
    """
    Get bottom level columns required to sum to create grouped columns.

    Parameters
    ----------
    grouped_levels : List[str]
        Grouped level, underscore delimited, column names.
    bottom_levels : List[str]
        Bottom level, underscore delimited, column names.

    Returns
    -------
    List[List[str]]
        Bottom level column names that make up each individual aggregated node in the hierarchy.
    """
    # Split groupings by "_" b/c this makes it possible to search column names
    grouped_levels_split = [lev.split("_") for lev in grouped_levels]
    bottom_levels_split = [lev.split("_") for lev in bottom_levels]

    cols_to_add = []
    for lev in grouped_levels_split:
        group_bl_cols = [
            bl_col for bl_col in bottom_levels_split if set(lev).issubset(bl_col)
        ]
        cols_to_add.append(["_".join(lev) for lev in group_bl_cols])
    return cols_to_add


def get_hierarchichal_df(
    df: pandas.DataFrame,
    level_names: List[str],
    hierarchy: List[List[str]],
    date_colname: str,
    val_colname: str,
) -> Tuple[pandas.DataFrame, np.array, List[str]]:
    """
    Transform your tabular dataframe to a wide dataframe with desired levels a hierarchy.

    Parameters
    ----------
    df : pd.DataFrame
        Tabular dataframe
    level_names : List[str]
        Levels in the hierarchy.
    hierarchy : List[List[str]]
        Desired levels in your hierarchy.
    date_colname : str
        Date column name
    val_colname : str
        Name of column containing series values.

    Returns
    -------
    pd.DataFrame
        Wide dataframe with levels of specified aggregation.

    np.array
        Summing matrix.

    List[str]:
        Summing matrix labels.

    Examples
    --------
    >>> import hts.functions
    >>> hier_df = pandas.DataFrame(
        data={
            'ds': ['2020-01', '2020-02'] * 5,
            "lev1": ['A', 'A',
                     'A', 'A',
                     'A', 'A',
                     'B', 'B',
                     'B', 'B'],
            "lev2": ['X', 'X',
                     'Y', 'Y',
                     'Z', 'Z',
                     'X', 'X',
                     'Y', 'Y'],
            "val": [1, 2,
                    3, 4,
                    5, 6,
                    7, 8,
                    9, 10]
        }
    )
    >>> hier_df
            ds lev1 lev2  val
    0  2020-01    A    X    1
    1  2020-02    A    X    2
    2  2020-01    A    Y    3
    3  2020-02    A    Y    4
    4  2020-01    A    Z    5
    5  2020-02    A    Z    6
    6  2020-01    B    X    7
    7  2020-02    B    X    8
    8  2020-01    B    Y    9
    9  2020-02    B    Y   10
    >>> level_names = ['lev1', 'lev2']
    >>> hierarchy = [['lev1'], ['lev2']]
    >>> wide_df, sum_mat, sum_mat_labels = hts.functions.get_hierarchichal_df(hier_df,
                                                                              level_names=level_names,
                                                                              hierarchy=hierarchy,
                                                                              date_colname='ds',
                                                                              val_colname='val')
    >>> wide_df
        lev1_lev2  A_X  A_Y  A_Z  B_X  B_Y  total   A   B   X   Y  Z
        ds
        2020-01      1    3    5    7    9     25   9  16   8  12  5
        2020-02      2    4    6    8   10     30  12  18  10  14  6
    """
    # Column names separated by underscores
    level_names_underscores = "_".join(level_names)

    # Create a column representing the bottom level of aggregation
    df[level_names_underscores] = _create_bl_str_col(df, level_names)

    # Pivot df to bottom level. We can create the aggregate these series to get all the higher levels.
    forecast_df = df.pivot(
        index=date_colname, columns=level_names_underscores, values=val_colname
    )

    # Sum all bottom level series to get total
    forecast_df["total"] = forecast_df.sum(axis=1)

    bottom_levels = list(df[level_names_underscores].unique())

    grouped_levels = get_agg_series(df, hierarchy)

    sum_mat, sum_mat_labels = to_sum_mat(
        ntree=None, node_labels=[["total"], grouped_levels, bottom_levels]
    )

    forecast_df = add_agg_series_to_df(forecast_df, grouped_levels, bottom_levels)

    return forecast_df, sum_mat, sum_mat_labels
